import csv
import json

import data_browser.models
import pytest
from django.contrib.auth.models import User

from . import models

true = True
false = False
null = None


class ANY:  # pragma: no cover
    def __init__(self, type):
        self.type = type

    def __eq__(self, other):
        return isinstance(other, self.type)


@pytest.fixture
def products(db):
    address = models.Address.objects.create(city="london")
    producer = models.Producer.objects.create(name="Bob", address=address)
    models.Product.objects.create(name="a", size=1, size_unit="g", producer=producer)
    models.Product.objects.create(name="b", size=1, size_unit="g", producer=producer)
    models.Product.objects.create(name="c", size=2, size_unit="g", producer=producer)


def test_query_html(admin_client):
    res = admin_client.get(
        "/data_browser/query/tests.Product/size-0,name+1,size_unit.html?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    context = json.loads(res.context["ctx"])
    assert context.keys() == {"config", "initialState", "sentryDsn"}

    assert context["initialState"].keys() == {"model", "fields", "filters", "data"}
    assert context["initialState"]["model"] == "tests.Product"
    print(json.dumps(context["initialState"]["fields"], indent=4, sort_keys=True))
    assert context["initialState"]["fields"] == [
        {"path": "size", "priority": 0, "sort": "dsc"},
        {"path": "name", "priority": 1, "sort": "asc"},
        {"path": "size_unit", "priority": null, "sort": null},
    ]
    assert context["initialState"]["filters"] == [
        {"errorMessage": None, "lookup": "lt", "path": "size", "value": "2"},
        {"errorMessage": None, "lookup": "gt", "path": "id", "value": "0"},
    ]
    assert context["initialState"]["data"] == []

    assert context["config"].keys() == {
        "baseUrl",
        "adminUrl",
        "sortedModels",
        "types",
        "allModelFields",
    }
    assert context["config"]["baseUrl"] == "/data_browser/"
    assert context["config"]["adminUrl"] == "/admin/data_browser/view/add/"
    assert context["config"]["sortedModels"] == [
        "auth.Group",
        "auth.User",
        "data_browser.View",
        "tests.Address",
        "tests.InAdmin",
        "tests.InlineAdmin",
        "tests.Normal",
        "tests.Producer",
        "tests.Product",
        "tests.SKU",
        "tests.Tag",
    ]
    print(json.dumps(context["config"]["types"], indent=4, sort_keys=True))
    assert context["config"]["types"] == {
        "boolean": {
            "defaultLookup": "equals",
            "defaultValue": true,
            "lookups": {
                "equals": {"type": "boolean"},
                "is_null": {"type": "boolean"},
                "not_equals": {"type": "boolean"},
            },
            "sortedLookups": ["equals", "not_equals", "is_null"],
        },
        "html": {
            "defaultLookup": null,
            "defaultValue": "",
            "lookups": {},
            "sortedLookups": [],
        },
        "number": {
            "defaultLookup": "equals",
            "defaultValue": 0,
            "lookups": {
                "equals": {"type": "number"},
                "gt": {"type": "number"},
                "gte": {"type": "number"},
                "is_null": {"type": "boolean"},
                "lt": {"type": "number"},
                "lte": {"type": "number"},
                "not_equals": {"type": "number"},
            },
            "sortedLookups": [
                "equals",
                "not_equals",
                "gt",
                "gte",
                "lt",
                "lte",
                "is_null",
            ],
        },
        "string": {
            "defaultLookup": "equals",
            "defaultValue": "",
            "lookups": {
                "contains": {"type": "string"},
                "ends_with": {"type": "string"},
                "equals": {"type": "string"},
                "is_null": {"type": "boolean"},
                "not_contains": {"type": "string"},
                "not_ends_with": {"type": "string"},
                "not_equals": {"type": "string"},
                "not_regex": {"type": "string"},
                "not_starts_with": {"type": "string"},
                "regex": {"type": "string"},
                "starts_with": {"type": "string"},
            },
            "sortedLookups": [
                "equals",
                "contains",
                "starts_with",
                "ends_with",
                "regex",
                "not_equals",
                "not_contains",
                "not_starts_with",
                "not_ends_with",
                "not_regex",
                "is_null",
            ],
        },
        "time": {
            "defaultLookup": "equals",
            "defaultValue": ANY(str),
            "lookups": {
                "equals": {"type": "time"},
                "gt": {"type": "time"},
                "gte": {"type": "time"},
                "is_null": {"type": "boolean"},
                "lt": {"type": "time"},
                "lte": {"type": "time"},
                "not_equals": {"type": "time"},
            },
            "sortedLookups": [
                "equals",
                "not_equals",
                "gt",
                "gte",
                "lt",
                "lte",
                "is_null",
            ],
        },
    }
    print(json.dumps(context["config"]["allModelFields"], indent=4, sort_keys=True))
    assert context["config"]["allModelFields"] == {
        "auth.Group": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {},
            "sorted_fields": ["id", "admin", "name"],
            "sorted_fks": [],
        },
        "auth.User": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "date_joined": {"concrete": true, "type": "time"},
                "email": {"concrete": true, "type": "string"},
                "first_name": {"concrete": true, "type": "string"},
                "id": {"concrete": true, "type": "number"},
                "is_active": {"concrete": true, "type": "boolean"},
                "is_staff": {"concrete": true, "type": "boolean"},
                "is_superuser": {"concrete": true, "type": "boolean"},
                "last_login": {"concrete": true, "type": "time"},
                "last_name": {"concrete": true, "type": "string"},
                "password": {"concrete": true, "type": "string"},
                "username": {"concrete": true, "type": "string"},
            },
            "fks": {},
            "sorted_fields": [
                "id",
                "admin",
                "date_joined",
                "email",
                "first_name",
                "is_active",
                "is_staff",
                "is_superuser",
                "last_login",
                "last_name",
                "password",
                "username",
            ],
            "sorted_fks": [],
        },
        "data_browser.View": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "created_time": {"concrete": true, "type": "time"},
                "description": {"concrete": true, "type": "string"},
                "fields": {"concrete": true, "type": "string"},
                "id": {"concrete": true, "type": "string"},
                "model_name": {"concrete": true, "type": "string"},
                "name": {"concrete": true, "type": "string"},
                "public": {"concrete": true, "type": "boolean"},
                "query": {"concrete": true, "type": "string"},
            },
            "fks": {"owner": {"model": "auth.User"}},
            "sorted_fields": [
                "id",
                "admin",
                "created_time",
                "description",
                "fields",
                "model_name",
                "name",
                "public",
                "query",
            ],
            "sorted_fks": ["owner"],
        },
        "tests.Address": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "city": {"concrete": true, "type": "string"},
                "id": {"concrete": true, "type": "number"},
            },
            "fks": {},
            "sorted_fields": ["id", "admin", "city"],
            "sorted_fks": [],
        },
        "tests.InAdmin": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {},
            "sorted_fields": ["id", "admin", "name"],
            "sorted_fks": [],
        },
        "tests.InlineAdmin": {
            "fields": {
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {"in_admin": {"model": "tests.InAdmin"}},
            "sorted_fields": ["id", "name"],
            "sorted_fks": ["in_admin"],
        },
        "tests.Normal": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {
                "in_admin": {"model": "tests.InAdmin"},
                "inline_admin": {"model": "tests.InlineAdmin"},
            },
            "sorted_fields": ["id", "admin", "name"],
            "sorted_fks": ["in_admin", "inline_admin"],
        },
        "tests.Producer": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {"address": {"model": "tests.Address"}},
            "sorted_fields": ["id", "admin", "name"],
            "sorted_fks": ["address"],
        },
        "tests.Product": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "is_onsale": {"concrete": false, "type": "string"},
                "name": {"concrete": true, "type": "string"},
                "onsale": {"concrete": true, "type": "boolean"},
                "size": {"concrete": true, "type": "number"},
                "size_unit": {"concrete": true, "type": "string"},
            },
            "fks": {
                "default_sku": {"model": "tests.SKU"},
                "producer": {"model": "tests.Producer"},
            },
            "sorted_fields": [
                "id",
                "admin",
                "is_onsale",
                "name",
                "onsale",
                "size",
                "size_unit",
            ],
            "sorted_fks": ["default_sku", "producer"],
        },
        "tests.SKU": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {"product": {"model": "tests.Product"}},
            "sorted_fields": ["id", "admin", "name"],
            "sorted_fks": ["product"],
        },
        "tests.Tag": {
            "fields": {
                "admin": {"concrete": false, "type": "html"},
                "id": {"concrete": true, "type": "number"},
                "name": {"concrete": true, "type": "string"},
            },
            "fks": {},
            "sorted_fields": ["id", "admin", "name"],
            "sorted_fks": [],
        },
    }


@pytest.mark.usefixtures("products")
def test_query_json_bad_fields(admin_client):
    res = admin_client.get(
        "".join(
            [
                "/data_browser/query/tests.Product/",
                "size-0,name+1,size_unit,bob-2,is_onsale,pooducer__name,producer__name.json",
                "?size__lt=2&id__gt=0&bob__gt=1&size__xx=1&size__lt=xx",
            ]
        )
    )
    assert res.status_code == 200
    assert json.loads(res.content.decode("utf-8"))["data"] == [
        [1, "a", "g", False, "Bob"],
        [1, "b", "g", False, "Bob"],
    ]


def test_query_html_bad_model(admin_client):
    res = admin_client.get(
        "/data_browser/query/tests.Bob/size-0,name+1,size_unit.html?size__lt=2&id__gt=0"
    )
    assert res.status_code == 404


@pytest.mark.usefixtures("products")
def test_query_csv(admin_client):
    res = admin_client.get(
        "/data_browser/query/tests.Product/size-0,name+1,size_unit.csv?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    rows = list(csv.reader(res.content.decode("utf-8").splitlines()))
    assert rows == [["size", "name", "size_unit"], ["1", "a", "g"], ["1", "b", "g"]]


@pytest.mark.usefixtures("products")
def test_query_json(admin_client):
    res = admin_client.get(
        "/data_browser/query/tests.Product/size-0,name+1,size_unit.json?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    data = json.loads(res.content.decode("utf-8"))

    print(json.dumps(data, indent=4, sort_keys=True))
    assert data == {
        "data": [[1, "a", "g"], [1, "b", "g"]],
        "fields": [
            {"path": "size", "priority": 0, "sort": "dsc"},
            {"path": "name", "priority": 1, "sort": "asc"},
            {"path": "size_unit", "priority": null, "sort": null},
        ],
        "filters": [
            {"errorMessage": null, "lookup": "lt", "path": "size", "value": "2"},
            {"errorMessage": null, "lookup": "gt", "path": "id", "value": "0"},
        ],
        "model": "tests.Product",
    }


@pytest.mark.usefixtures("products")
def test_query_json_bad_model(admin_client):
    res = admin_client.get(
        "/data_browser/query/tests.Bob/size-0,name+1,size_unit.json?size__lt=2&id__gt=0"
    )
    assert res.status_code == 404


@pytest.mark.usefixtures("products")
def test_view_csv(admin_client):
    view = data_browser.models.View.objects.create(
        model_name="tests.Product",
        fields="size-0,name+1,size_unit",
        query="size__lt=2&id__gt=0",
        owner=User.objects.get(),
    )

    res = admin_client.get(f"/data_browser/view/{view.pk}.csv")
    assert res.status_code == 404

    view.public = True
    view.save()
    res = admin_client.get(f"/data_browser/view/{view.pk}.csv")
    assert res.status_code == 200
    rows = list(csv.reader(res.content.decode("utf-8").splitlines()))
    assert rows == [["size", "name", "size_unit"], ["1", "a", "g"], ["1", "b", "g"]]


@pytest.mark.usefixtures("products")
def test_view_json(admin_client):
    view = data_browser.models.View.objects.create(
        model_name="tests.Product",
        fields="size-0,name+1,size_unit",
        query="size__lt=2&id__gt=0",
        owner=User.objects.get(),
    )

    res = admin_client.get(f"/data_browser/view/{view.pk}.json")
    assert res.status_code == 404

    view.public = True
    view.save()
    res = admin_client.get(f"/data_browser/view/{view.pk}.json")
    assert res.status_code == 200
    data = json.loads(res.content.decode("utf-8"))
    print(json.dumps(data, indent=4, sort_keys=True))
    assert data == {
        "data": [[1, "a", "g"], [1, "b", "g"]],
        "fields": [
            {"path": "size", "priority": 0, "sort": "dsc"},
            {"path": "name", "priority": 1, "sort": "asc"},
            {"path": "size_unit", "priority": null, "sort": null},
        ],
        "filters": [
            {"errorMessage": null, "lookup": "lt", "path": "size", "value": "2"},
            {"errorMessage": null, "lookup": "gt", "path": "id", "value": "0"},
        ],
        "model": "tests.Product",
    }



# Generated by Django 2.0.13 on 2020-03-31 17:42

import data_browser.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="view",
            name="id",
            field=models.CharField(
                default=data_browser.models.get_id,
                max_length=12,
                primary_key=True,
                serialize=False,
            ),
        ),
    ]



from django.conf import settings

settings.configure(
    INSTALLED_APPS=[
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.staticfiles",
        "django.contrib.admin",
        "tests",
        "data_browser",
    ],
    DATABASES={
        "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "db.sqlite3"}
    },
    ROOT_URLCONF="tests.urls",
    MIDDLEWARE=[
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
    ],
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "OPTIONS": {
                "context_processors": [
                    "django.template.context_processors.debug",
                    "django.template.context_processors.request",
                    "django.template.context_processors.static",
                    "django.contrib.auth.context_processors.auth",
                ],
                "loaders": ["django.template.loaders.app_directories.Loader"],
            },
        }
    ],
    STATIC_URL="/static/",
)



import os

from django.conf import settings
from django.urls import path, register_converter
from django.views.generic.base import RedirectView
from django.views.static import serve

from .views import proxy_js_dev_server, query, query_ctx, query_html, view

FE_BUILD_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "fe_build")


class Optional:
    regex = "[^/]*"

    def to_python(self, value):
        return value

    def to_url(self, value):
        return value


register_converter(Optional, "optional")

app_name = "data_browser"

urlpatterns = [
    path(
        "query/<optional:model_name>/<optional:fields>.html",
        query_html,
        name="query_html",
    ),
    path(
        "query/<optional:model_name>/<optional:fields>.ctx", query_ctx, name="query_ctx"
    ),
    path("query/<optional:model_name>/<optional:fields>.<media>", query, name="query"),
    path("view/<pk>.<media>", view, name="view"),
    path("", RedirectView.as_view(url="query//.html?", permanent=False), name="home"),
]
if getattr(settings, "DATA_BROWSER_DEV", False):  # pragma: no cover
    urlpatterns.append(path("<path:path>", proxy_js_dev_server))
else:
    urlpatterns.append(path("<path:path>", serve, {"document_root": FE_BUILD_DIR}))



# Generated by Django 2.0.13 on 2020-04-29 19:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0002_auto_20200331_1842"),
    ]

    operations = [
        migrations.RemoveField(model_name="view", name="app",),
    ]



# Generated by Django 2.0.13 on 2020-03-31 16:35

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="View",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_time",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("name", models.CharField(max_length=64)),
                ("description", models.TextField(blank=True)),
                ("public", models.BooleanField(default=False)),
                ("app", models.CharField(max_length=16)),
                ("model", models.CharField(max_length=32)),
                ("fields", models.TextField(blank=True)),
                ("query", models.TextField()),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



import pytest
from data_browser import orm
from data_browser.query import (
    ASC,
    DSC,
    BooleanFieldType,
    BoundFilter,
    BoundQuery,
    NumberFieldType,
    Query,
    QueryField,
    QueryFilter,
    StringFieldType,
    TimeFieldType,
)


@pytest.fixture
def query():
    return Query(
        "app.model",
        [QueryField("fa", ASC, 1), QueryField("fd", DSC, 0), QueryField("fn")],
        [QueryFilter("bob", "equals", "fred")],
    )


@pytest.fixture
def bound_query(query):
    orm_models = {
        "app.model": orm.OrmModel(
            fields={
                "fa": orm.OrmField(type_=StringFieldType, concrete=True),
                "fd": orm.OrmField(type_=StringFieldType, concrete=True),
                "fn": orm.OrmField(type_=StringFieldType, concrete=False),
                "bob": orm.OrmField(type_=StringFieldType, concrete=True),
            },
            fks={"tom": "app.Tom"},
        ),
        "app.Tom": orm.OrmModel(
            fields={"jones": orm.OrmField(type_=StringFieldType, concrete=True)},
            fks={"michael": "app.Michael"},
        ),
        "app.Michael": orm.OrmModel(
            fields={"bolton": orm.OrmField(type_=StringFieldType, concrete=True)},
            fks={},
        ),
    }
    return BoundQuery(query, orm_models)


class TestQuery:
    def test_from_request(self, query):
        q = Query.from_request("app.model", "fa+1,fd-0,fn", {"bob__equals": ["fred"]})
        assert q == query

    def test_from_request_with_related_filter(self):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", {"bob__jones__equals": ["fred"]}
        )
        assert q == Query(
            "app.model",
            [QueryField("fa", ASC, 1), QueryField("fd", DSC, 0), QueryField("fn")],
            [QueryFilter("bob__jones", "equals", "fred")],
        )

    def test_url(self, query):
        assert (
            query.get_url("html")
            == "/data_browser/query/app.model/fa+1,fd-0,fn.html?bob__equals=fred"
        )

    def test_url_no_filters(self, query):
        query.filters = []
        assert (
            query.get_url("html") == "/data_browser/query/app.model/fa+1,fd-0,fn.html?"
        )


class TestBoundQuery:
    def test_fields(self, bound_query):
        assert [f.path for f in bound_query.fields] == ["fa", "fd", "fn"]

    def test_calculated_fields(self, bound_query):
        assert bound_query.calculated_fields == {"fn"}

    def test_sort_fields(self, bound_query):
        assert [(f.path, f.direction, f.priority) for f in bound_query.sort_fields] == [
            ("fd", DSC, 0),
            ("fa", ASC, 1),
        ]

    def test_filters(self, bound_query):
        assert bound_query.filters == [
            BoundFilter("bob", "equals", "fred", StringFieldType)
        ]


class TestFieldType:
    def test_repr(self):
        assert repr(StringFieldType) == f"StringFieldType"


class TestStringFieldType:
    def test_validate(self):
        assert BoundFilter("bob", "contains", "hello", StringFieldType).is_valid
        assert not BoundFilter("bob", "pontains", "hello", StringFieldType).is_valid

    def test_default_lookup(self):
        assert StringFieldType.default_lookup == "equals"


class TestNumberFieldType:
    def test_validate(self):
        assert BoundFilter("bob", "gt", "6.1", NumberFieldType).is_valid
        assert not BoundFilter("bob", "pontains", "6.1", NumberFieldType).is_valid
        assert not BoundFilter("bob", "gt", "hello", NumberFieldType).is_valid
        assert BoundFilter("bob", "is_null", "True", NumberFieldType).is_valid
        assert not BoundFilter("bob", "is_null", "hello", NumberFieldType).is_valid

    def test_default_lookup(self):
        assert NumberFieldType.default_lookup == "equals"


class TestTimeFieldType:
    def test_validate(self):
        assert BoundFilter("bob", "gt", "2018-03-20T22:31:23", TimeFieldType).is_valid
        assert not BoundFilter("bob", "gt", "hello", TimeFieldType).is_valid
        assert not BoundFilter(
            "bob", "pontains", "2018-03-20T22:31:23", TimeFieldType
        ).is_valid
        assert BoundFilter("bob", "is_null", "True", TimeFieldType).is_valid
        assert not BoundFilter("bob", "is_null", "hello", TimeFieldType).is_valid

    def test_default_lookup(self):
        assert TimeFieldType.default_lookup == "equals"


class TestBooleanFieldType:
    def test_validate(self):
        assert BoundFilter("bob", "equals", "True", BooleanFieldType).is_valid
        assert not BoundFilter("bob", "equals", "hello", BooleanFieldType).is_valid
        assert not BoundFilter("bob", "pontains", "True", BooleanFieldType).is_valid

    def test_default_lookup(self):
        assert BooleanFieldType.default_lookup == "equals"



import csv
import io
import json
import sys

import django.contrib.admin.views.decorators as admin_decorators
from django import http
from django.conf import settings
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404
from django.template import engines, loader
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt

from .models import View
from .orm import _OPEN_IN_ADMIN, get_data, get_models
from .query import TYPES, BoundQuery, Query


def _get_query_data(bound_query):
    return {
        "filters": [
            {
                "errorMessage": filter_.err_message,
                "path": filter_.path,
                "lookup": filter_.lookup,
                "value": filter_.value,
            }
            for filter_ in bound_query.filters
        ],
        "fields": [
            {"path": field.path, "sort": field.direction, "priority": field.priority}
            for field in bound_query.fields
        ],
        "model": bound_query.model_name,
    }


def _get_config(orm_models):
    types = {
        name: {
            "lookups": {n: {"type": t} for n, t in type_.lookups.items()},
            "sortedLookups": list(type_.lookups),
            "defaultLookup": type_.default_lookup,
            "defaultValue": type_.default_value,
        }
        for name, type_ in TYPES.items()
    }

    def sort_model_fields(fields):
        fields = sorted(fields)
        front = {"id": 1, _OPEN_IN_ADMIN: 2}
        return sorted(fields, key=lambda x: front.get(x, sys.maxsize))

    orm_models = {
        model_name: {
            "fields": {
                name: {"type": orm_field.type_.name, "concrete": orm_field.concrete}
                for name, orm_field in orm_model.fields.items()
            },
            "fks": {
                name: {"model": fk_field.model_name}
                for name, fk_field in orm_model.fks.items()
            },
            "sorted_fields": sort_model_fields(orm_model.fields),
            "sorted_fks": sorted(orm_model.fks),
        }
        for model_name, orm_model in orm_models.items()
    }

    return {
        "baseUrl": reverse("data_browser:home"),
        "adminUrl": reverse(f"admin:{View._meta.db_table}_add"),
        "types": types,
        "allModelFields": orm_models,
        "sortedModels": sorted(orm_models),
    }


def _get_context(request, model_name, fields):
    query = Query.from_request(model_name, fields, request.GET)
    orm_models = get_models(request)
    if query.model_name and query.model_name not in orm_models:
        raise http.Http404(f"query.model_name does not exist")
    bound_query = BoundQuery(query, orm_models)
    return {
        "config": _get_config(orm_models),
        "initialState": {"data": [], **_get_query_data(bound_query)},
        "sentryDsn": getattr(settings, "DATA_BROWSER_FE_DSN", None),
    }


@admin_decorators.staff_member_required
def query_ctx(request, *, model_name, fields=""):  # pragma: no cover
    ctx = _get_context(request, model_name, fields)
    return JsonResponse(ctx)


@admin_decorators.staff_member_required
def query_html(request, *, model_name="", fields=""):
    ctx = _get_context(request, model_name, fields)
    ctx = json.dumps(ctx)
    ctx = ctx.replace("<", "\\u003C").replace(">", "\\u003E").replace("&", "\\u0026")

    if getattr(settings, "DATA_BROWSER_DEV", False):  # pragma: no cover
        response = _get_from_js_dev_server(request)
        template = engines["django"].from_string(response.text)
    else:
        template = loader.get_template("data_browser/index.html")

    return TemplateResponse(request, template, {"ctx": ctx})


@admin_decorators.staff_member_required
def query(request, *, model_name, fields="", media):
    query = Query.from_request(model_name, fields, request.GET)
    return _data_response(request, query, media)


def view(request, pk, media):
    view = get_object_or_404(View.objects.filter(public=True), pk=pk)
    request.user = view.owner  # public views are run as the person who owns them
    query = view.get_query()
    return _data_response(request, query, media)


def _data_response(request, query, media):
    orm_models = get_models(request)
    if query.model_name not in orm_models:
        raise http.Http404(f"query.model_name does not exist")
    bound_query = BoundQuery(query, orm_models)
    data = get_data(request, bound_query)

    if media == "csv":
        buffer = io.StringIO()
        writer = csv.writer(buffer)
        writer.writerow(f.path for f in bound_query.fields)
        writer.writerows(data)
        buffer.seek(0)
        response = HttpResponse(buffer, content_type="text/csv")
        response[
            "Content-Disposition"
        ] = f"attachment; filename={query.model_name}-{timezone.now().isoformat()}.csv"
        return response
    elif media == "json":
        resp = _get_query_data(bound_query)
        resp["data"] = data
        return JsonResponse(resp)
    else:
        assert False


def _get_from_js_dev_server(request):  # pragma: no cover
    import requests

    upstream_url = f"http://localhost:3000{request.path}"
    method = request.META["REQUEST_METHOD"].lower()
    return getattr(requests, method)(upstream_url, stream=True)


@csrf_exempt
def proxy_js_dev_server(request, path):  # pragma: no cover
    """
    Proxy HTTP requests to the frontend dev server in development.

    The implementation is very basic e.g. it doesn't handle HTTP headers.

    """
    response = _get_from_js_dev_server(request)
    content_type = response.headers.get("Content-Type")

    if request.META.get("HTTP_UPGRADE", "").lower() == "websocket":
        return http.HttpResponse(
            content="WebSocket connections aren't supported",
            status=501,
            reason="Not Implemented",
        )

    else:
        return http.StreamingHttpResponse(
            streaming_content=response.iter_content(2 ** 12),
            content_type=content_type,
            status=response.status_code,
            reason=response.reason,
        )



from collections import defaultdict
from dataclasses import dataclass

from django.contrib import admin
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.admin.utils import flatten_fieldsets
from django.db import models
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.forms.models import _get_foreign_key
from django.urls import reverse

from .query import (
    ASC,
    DSC,
    BooleanFieldType,
    FieldType,
    HTMLFieldType,
    NumberFieldType,
    StringFieldType,
    TimeFieldType,
)

_OPEN_IN_ADMIN = "admin"

_FIELD_MAP = [
    ((models.BooleanField, models.NullBooleanField), BooleanFieldType),
    (
        (
            models.CharField,
            models.TextField,
            models.GenericIPAddressField,
            models.UUIDField,
        ),
        StringFieldType,
    ),
    ((models.DateTimeField, models.DateField), TimeFieldType),
    (
        (models.DecimalField, models.FloatField, models.IntegerField, models.AutoField),
        NumberFieldType,
    ),
    ((models.FileField,), None),
]


def get_model_name(model, sep="."):
    return f"{model._meta.app_label}{sep}{model.__name__}"


@dataclass
class OrmModel:
    fields: dict
    fks: dict
    admin: BaseModelAdmin = None


@dataclass
class OrmField:
    type_: FieldType
    concrete: bool


@dataclass
class OrmFkField:
    model_name: str


def _get_all_admin_fields(request):
    request.data_browser = True

    def from_fieldsets(admin, model):
        obj = model()  # we want the admin change field sets, not the add ones
        for f in flatten_fieldsets(admin.get_fieldsets(request, obj)):
            if hasattr(model, f):
                yield f

    def visible(modeladmin, request):
        if modeladmin.has_change_permission(request):
            return True
        if hasattr(modeladmin, "has_view_permission"):
            return modeladmin.has_view_permission(request)
        return False

    all_admin_fields = defaultdict(set)
    model_admins = {}
    for model, modeladmin in admin.site._registry.items():
        model_admins[model] = modeladmin
        if visible(modeladmin, request):
            all_admin_fields[model].update(from_fieldsets(modeladmin, model))
            all_admin_fields[model].add(_OPEN_IN_ADMIN)

            # check the inlines, these are already filtered for access
            for inline in modeladmin.get_inline_instances(request):
                if inline.model not in model_admins:  # pragma: no cover
                    model_admins[inline.model] = inline
                all_admin_fields[inline.model].update(
                    from_fieldsets(inline, inline.model)
                )
                all_admin_fields[inline.model].add(
                    _get_foreign_key(model, inline.model, inline.fk_name).name
                )

    # we always have id and never pk
    for fields in all_admin_fields.values():
        fields.add("id")
        fields.discard("pk")

    return model_admins, all_admin_fields


def _get_fields_for_model(model, model_admins, admin_fields):
    fields = {}
    fks = {}

    model_fields = {f.name: f for f in model._meta.get_fields()}

    for field_name in admin_fields[model]:
        field = model_fields.get(field_name)
        if field_name == _OPEN_IN_ADMIN:
            fields[_OPEN_IN_ADMIN] = OrmField(type_=HTMLFieldType, concrete=False)
        elif isinstance(field, (ForeignObjectRel, models.ManyToManyField)):
            pass  # TODO 2many support
        elif isinstance(field, models.ForeignKey):
            if field.related_model in admin_fields:
                fks[field_name] = OrmFkField(get_model_name(field.related_model))
        elif isinstance(field, type(None)):
            fields[field_name] = OrmField(type_=StringFieldType, concrete=False)
        else:
            for django_types, field_type in _FIELD_MAP:
                if isinstance(field, django_types):
                    if field_type:
                        fields[field_name] = OrmField(type_=field_type, concrete=True)
                    break
            else:  # pragma: no cover
                print(
                    f"DataBrowser: {model.__name__}.{field_name} unknown type {type(field).__name__}"
                )

    return OrmModel(fields=fields, fks=fks, admin=model_admins[model])


def get_models(request):
    model_admins, admin_fields = _get_all_admin_fields(request)
    # {model: {"fields": {field_name, FieldType}, "fks": {field_name: model}}}
    return {
        get_model_name(model): _get_fields_for_model(model, model_admins, admin_fields)
        for model in admin_fields
    }


def _get_django_lookup(field_type, lookup):
    if field_type == StringFieldType and lookup == "equals":
        return "iexact"
    else:
        lookup = {
            "equals": "exact",
            "regex": "iregex",
            "contains": "icontains",
            "starts_with": "istartswith",
            "ends_with": "iendswith",
            "is_null": "isnull",
            "gt": "gt",
            "gte": "gte",
            "lt": "lt",
            "lte": "lte",
        }[lookup]
        return lookup


def get_data(request, bound_query):
    request.data_browser = True

    if not bound_query.fields:
        return []

    admin = bound_query.orm_models[bound_query.model_name].admin
    qs = admin.get_queryset(request)

    # sort
    sort_fields = []
    for field in bound_query.sort_fields:
        if field.direction is ASC:
            sort_fields.append(field.path)
        if field.direction is DSC:
            sort_fields.append(f"-{field.path}")
    qs = qs.order_by(*sort_fields)

    # filter
    for filter_ in bound_query.filters:
        if filter_.is_valid:
            negation = False

            lookup = filter_.lookup
            if lookup.startswith("not_"):
                negation = True
                lookup = lookup[4:]

            filter_str = f"{filter_.path}__{_get_django_lookup(filter_.type_, lookup)}"
            if negation:
                qs = qs.exclude(**{filter_str: filter_.parsed})
            else:
                qs = qs.filter(**{filter_str: filter_.parsed})

    # no calculated fields early out using qs.values
    if not bound_query.calculated_fields:
        data = []
        for row in qs.values(*[f.path for f in bound_query.fields]).distinct():
            data.append(
                [field.type_.format(row[field.path]) for field in bound_query.fields]
            )
        return data

    # preloading
    select_related = set()

    def add_select_relateds(path):
        while "__" in path:
            path = path.rsplit("__", 1)[0]
            select_related.add(path)

    for field in bound_query.sort_fields:
        add_select_relateds(field.path)

    for filter_ in bound_query.filters:
        if filter_.is_valid:
            add_select_relateds(filter_.path)

    prefetch_related = set()
    for field in bound_query.fields:
        if "__" in field.path:
            prefetch_related.add(field.path.rsplit("__", 1)[0])
    prefetch_related -= select_related

    if select_related:
        qs = qs.select_related(*select_related)
    if prefetch_related:
        qs = qs.prefetch_related(*prefetch_related)

    def get_admin_link(obj):
        if obj is None:
            return "null"
        model_name = get_model_name(obj.__class__, "_")
        url_name = f"admin:{model_name}_change".lower()
        url = reverse(url_name, args=[obj.pk])
        return f'<a href="{url}">{obj}</a>'

    # get data
    def lookup(obj, path):
        value = obj
        for part in path.split("__"):
            if part == _OPEN_IN_ADMIN:
                value = get_admin_link(value)
            else:
                value = getattr(value, part, None)
        return value() if callable(value) else value

    data = []
    for row in qs.distinct():
        data.append(
            [
                field.type_.format(lookup(row, field.path))
                for field in bound_query.fields
            ]
        )
    return data



# Generated by Django 2.0.13 on 2020-05-01 08:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0003_remove_view_app"),
    ]

    operations = [
        migrations.RenameField(
            model_name="view", old_name="model", new_name="model_name",
        ),
    ]



import pytest
from data_browser.admin import ViewAdmin, globals
from data_browser.models import View
from django.contrib.auth.models import User


@pytest.fixture
def view():
    return View(model_name="app.model", fields="fa+0,fd-1,fn", query="bob__equals=fred")


def test_open_view(view, rf):
    expected = '<a href="/data_browser/query/app.model/fa+0,fd-1,fn.html?bob__equals=fred">view</a>'
    assert ViewAdmin.open_view(view) == expected


def test_public_link(view, rf):
    globals.request = rf.get("/")
    assert ViewAdmin.public_link(view) == "N/A"
    view.public = True
    expected = f"http://testserver/data_browser/view/{view.pk}.csv"
    assert ViewAdmin.public_link(view) == expected


def test_google_sheets_formula(view, rf):
    globals.request = rf.get("/")
    assert ViewAdmin.google_sheets_formula(view) == "N/A"
    view.public = True
    expected = f'=importdata("http://testserver/data_browser/view/{view.pk}.csv")'
    assert ViewAdmin.google_sheets_formula(view) == expected


def test_add_has_user(admin_client):
    user = User.objects.get()
    res = admin_client.get("/admin/data_browser/view/add/")
    assert res.status_code == 200
    assert res.context[0]["adminform"].form.initial["owner"] == user.pk


def test_change_form_links_have_full_url(view, admin_client):
    view.public = True
    view.save()
    res = admin_client.get(
        f"http://testserver/admin/data_browser/view/{view.pk}/change/"
    )
    assert res.status_code == 200

    all_fields = [f for fs in res.context[0]["adminform"] for l in fs for f in l]
    for field in all_fields:
        if field.is_readonly and field.field["name"] == "public_link":
            expected = f"http://testserver/data_browser/view/{view.pk}.csv"
            assert field.contents() == expected
            break
    else:
        assert False



import urllib
from dataclasses import dataclass
from typing import Optional, Sequence

import dateutil.parser
from django.urls import reverse
from django.utils import timezone

ASC, DSC = "asc", "dsc"


@dataclass
class QueryField:
    path: str
    direction: str = None
    priority: int = None

    def __post_init__(self):
        assert (self.direction is None) == (self.priority is None)


@dataclass
class QueryFilter:
    path: str
    lookup: str
    value: str


@dataclass
class Query:
    model_name: str
    fields: Sequence[QueryField]
    filters: Sequence[QueryFilter]

    @classmethod
    def from_request(cls, model_name, field_str, get_args):
        fields = []
        for field in field_str.split(","):
            field = field.strip()
            if field:
                if "+" in field:
                    path, priority = field.split("+")
                    fields.append(QueryField(path, ASC, int(priority)))
                elif "-" in field:
                    path, priority = field.split("-")
                    fields.append(QueryField(path, DSC, int(priority)))
                else:
                    fields.append(QueryField(field, None, None))

        filters = []
        for path__lookup, values in dict(get_args).items():
            for value in values:
                path, lookup = path__lookup.rsplit("__", 1)
                filters.append(QueryFilter(path, lookup, value))

        res = cls(model_name, fields, filters)
        return res

    @property
    def field_str(self):
        field_strs = []
        for field in self.fields:
            direction = {ASC: "+", DSC: "-", None: ""}[field.direction]
            priority = str(field.priority) if field.direction else ""
            field_strs.append(f"{field.path}{direction}{priority}")
        return ",".join(field_strs)

    @property
    def filter_fields(self):
        return [
            (f"{filter.path}__{filter.lookup}", filter.value) for filter in self.filters
        ]

    def get_url(self, media):
        base_url = reverse(
            "data_browser:query",
            kwargs={
                "model_name": self.model_name,
                "fields": self.field_str,
                "media": media,
            },
        )
        params = urllib.parse.urlencode(
            self.filter_fields, quote_via=urllib.parse.quote_plus, doseq=True
        )
        return f"{base_url}?{params}"


class MetaFieldType(type):
    def __repr__(cls):
        return cls.__name__

    @property
    def default_lookup(cls):
        return list(cls.lookups)[0] if cls.lookups else None

    @property
    def name(cls):
        name = cls.__name__.lower()
        assert name.endswith("fieldtype")
        return name[: -len("fieldtype")]


class FieldType(metaclass=MetaFieldType):
    default_value = ""

    def __init__(self):
        assert False

    @staticmethod
    def format(value):
        return value

    @staticmethod
    def parse(value):
        return value


class StringFieldType(FieldType):
    lookups = {
        "equals": "string",
        "contains": "string",
        "starts_with": "string",
        "ends_with": "string",
        "regex": "string",
        "not_equals": "string",
        "not_contains": "string",
        "not_starts_with": "string",
        "not_ends_with": "string",
        "not_regex": "string",
        "is_null": "boolean",
    }


class NumberFieldType(FieldType):
    default_value = 0
    lookups = {
        "equals": "number",
        "not_equals": "number",
        "gt": "number",
        "gte": "number",
        "lt": "number",
        "lte": "number",
        "is_null": "boolean",
    }

    @staticmethod
    def parse(value):
        return float(value)


class TimeFieldType(FieldType):
    lookups = {
        "equals": "time",
        "not_equals": "time",
        "gt": "time",
        "gte": "time",
        "lt": "time",
        "lte": "time",
        "is_null": "boolean",
    }
    default_value = timezone.now().strftime("%Y-%m-%d %H:%M")

    @staticmethod
    def parse(value):
        return timezone.make_aware(dateutil.parser.parse(value))

    @staticmethod
    def format(value):
        return timezone.make_naive(value) if value else None


class HTMLFieldType(FieldType):
    lookups = {}


class BooleanFieldType(FieldType):
    default_value = True
    lookups = {"equals": "boolean", "not_equals": "boolean", "is_null": "boolean"}

    @staticmethod
    def parse(value):
        value = value.lower()
        if value == "true":
            return True
        elif value == "false":
            return False
        else:
            raise ValueError("Expected 'true' or 'false'")


TYPES = {
    field_type.name: field_type
    for field_type in [
        StringFieldType,
        NumberFieldType,
        TimeFieldType,
        BooleanFieldType,
        HTMLFieldType,
    ]
}


@dataclass
class BoundFilter:
    path: str
    lookup: str
    value: str
    type_: FieldType

    @classmethod
    def bind(cls, query_filter, orm_field):
        return cls(
            query_filter.path, query_filter.lookup, query_filter.value, orm_field.type_
        )

    def __post_init__(self):
        self.parsed = None
        self.err_message = None

        lookups = self.type_.lookups
        if self.lookup not in lookups:
            self.err_message = f"Bad lookup '{self.lookup}' expected {lookups}"
        else:
            try:
                self.parsed = TYPES[lookups[self.lookup]].parse(self.value)
            except Exception as e:
                self.err_message = str(e) if str(e) else repr(e)

        self.is_valid = not self.err_message


@dataclass
class BoundField:
    path: str
    direction: Optional[str]
    priority: Optional[int]
    concrete: bool
    type_: FieldType

    @classmethod
    def bind(cls, query_field, orm_field):
        direction = query_field.direction if orm_field.concrete else None
        priority = query_field.priority if orm_field.concrete else None
        return cls(
            query_field.path, direction, priority, orm_field.concrete, orm_field.type_
        )


class BoundQuery:
    def __init__(self, query, orm_models):
        # orm_models = {model_name: {"fields": {field_name, FieldType}, "fks": {field_name: model_name}}}
        def get_orm_field(path):
            parts = path.split("__")
            model_name = query.model_name
            for part in parts[:-1]:
                fk_field = orm_models[model_name].fks.get(part)
                if fk_field is None:
                    return None
                model_name = fk_field.model_name
            res = orm_models[model_name].fields.get(parts[-1])
            return res

        self.model_name = query.model_name
        self.orm_models = orm_models

        self.fields = []
        for query_field in query.fields:
            orm_field = get_orm_field(query_field.path)
            if orm_field:
                self.fields.append(BoundField.bind(query_field, orm_field))

        self.filters = []
        for query_filter in query.filters:
            orm_field = get_orm_field(query_filter.path)
            if orm_field:
                self.filters.append(BoundFilter.bind(query_filter, orm_field))

    @property
    def sort_fields(self):
        return sorted((f for f in self.fields if f.direction), key=lambda f: f.priority)

    @property
    def calculated_fields(self):
        return {f.path for f in self.fields if not f.concrete}



import threading

from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html

from . import models

globals = threading.local()


@admin.register(models.View)
class ViewAdmin(admin.ModelAdmin):
    fieldsets = [
        (
            None,
            {
                "fields": [
                    "name",
                    "owner",
                    "public",
                    "open_view",
                    "public_link",
                    "google_sheets_formula",
                    "description",
                ]
            },
        ),
        ("Query", {"fields": ["model_name", "fields", "query"]}),
        ("Internal", {"fields": ["id", "created_time"]}),
    ]
    readonly_fields = ["open_view", "public_link", "google_sheets_formula", "id"]
    list_display = ["__str__", "owner", "public"]

    def change_view(self, request, *args, **kwargs):
        globals.request = request
        return super().change_view(request, *args, **kwargs)

    @staticmethod
    def open_view(obj):
        if not obj.model_name:
            return "N/A"
        url = obj.get_query().get_url("html")
        return format_html(f'<a href="{url}">view</a>')

    @staticmethod
    def public_link(obj):
        if obj.public:
            url = reverse("data_browser:view", kwargs={"pk": obj.pk, "media": "csv"})
            return globals.request.build_absolute_uri(url)
        else:
            return "N/A"

    @staticmethod
    def google_sheets_formula(obj):
        if obj.public:
            url = reverse("data_browser:view", kwargs={"pk": obj.pk, "media": "csv"})
            url = globals.request.build_absolute_uri(url)
            return f'=importdata("{url}")'
        else:
            return "N/A"

    def get_changeform_initial_data(self, request):
        get_data = super().get_changeform_initial_data(request)
        get_data["owner"] = request.user.pk
        return get_data



from django.contrib import admin

from . import models

# admin for perm testing


class InlineAdminInline(admin.TabularInline):
    model = models.InlineAdmin
    fields = ["name"]


@admin.register(models.InAdmin)
class InAdmin(admin.ModelAdmin):
    fields = ["name"]
    inlines = [InlineAdminInline]


@admin.register(models.Normal)
class Normal(admin.ModelAdmin):
    fields = ["name", "in_admin", "not_in_admin", "inline_admin"]


# general admin


@admin.register(models.Tag)
class Tag(admin.ModelAdmin):
    fields = ["name"]


@admin.register(models.Address)
class Address(admin.ModelAdmin):
    fields = ["pk", "city"]


class ProductMixin:
    fields = [
        "id",
        "producer",
        "name",
        "size",
        "size_unit",
        "producer",
        "is_onsale",
        "default_sku",
        "tags",
        "model_not_in_admin",
        "onsale",
    ]
    readonly_fields = ["is_onsale"]


class ProductInline(ProductMixin, admin.TabularInline):
    model = models.Product


@admin.register(models.Producer)
class Producer(admin.ModelAdmin):
    fields = ["name", "address"]
    inlines = [ProductInline]


class SKUMixin:
    fields = ["name", "product"]


class SKUInline(SKUMixin, admin.TabularInline):
    model = models.SKU


@admin.register(models.Product)
class Product(ProductMixin, admin.ModelAdmin):
    inlines = [SKUInline]


@admin.register(models.SKU)
class SKU(SKUMixin, admin.ModelAdmin):
    pass



from django.conf import settings
from django.db import models
from django.http import QueryDict
from django.utils import crypto, timezone


def get_id():
    return crypto.get_random_string(length=12)


class View(models.Model):
    id = models.CharField(primary_key=True, max_length=12, default=get_id)
    created_time = models.DateTimeField(default=timezone.now)

    name = models.CharField(max_length=64, blank=False)
    description = models.TextField(blank=True)
    public = models.BooleanField(default=False)
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.SET_NULL
    )

    model_name = models.CharField(max_length=32, blank=False)
    fields = models.TextField(blank=True)
    query = models.TextField(blank=False)

    def get_query(self):
        from .query import Query

        return Query.from_request(self.model_name, self.fields, QueryDict(self.query))

    def __str__(self):
        return f"{self.model_name} view: {self.name}"



import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


data_files = []
for directory in "data_browser/fe_build", "data_browser/templates":
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            data_files.append(os.path.join("..", path, filename))


setuptools.setup(
    name="django-data-browser",
    version="0.0.8",
    author="Gordon Wrigley",
    author_email="gordon.wrigley@gmail.com",
    description="git@github.com:tolomea/django-data-browser.git",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tolomea/django-data-browser",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
    package_data={"": data_files},
)



import pytest
from data_browser.models import View


@pytest.fixture
def view():
    return View(model_name="app.model", fields="+fa,-fd,fn", query="bob__equals=fred")


def test_str(view):
    view.name = "bob"
    assert str(view) == "app.model view: bob"



from django.db import models

# models for perm testing


class InAdmin(models.Model):
    name = models.TextField()


class NotInAdmin(models.Model):
    name = models.TextField()


class InlineAdmin(models.Model):
    name = models.TextField()
    in_admin = models.ForeignKey(InAdmin, on_delete=models.CASCADE)


class Normal(models.Model):
    name = models.TextField()
    in_admin = models.ForeignKey(InAdmin, on_delete=models.CASCADE)
    not_in_admin = models.ForeignKey(NotInAdmin, on_delete=models.CASCADE)
    inline_admin = models.ForeignKey(InlineAdmin, on_delete=models.CASCADE)


# general models


class Tag(models.Model):
    name = models.TextField()


class Address(models.Model):
    city = models.TextField()


class Producer(models.Model):
    address = models.ForeignKey(Address, on_delete=models.CASCADE)
    name = models.TextField()


class Product(models.Model):
    name = models.TextField()
    producer = models.ForeignKey(Producer, on_delete=models.CASCADE)
    size = models.IntegerField()
    size_unit = models.TextField()
    default_sku = models.ForeignKey("SKU", null=True, on_delete=models.CASCADE)
    tags = models.ManyToManyField(Tag)
    onsale = models.BooleanField(null=True)

    not_in_admin = models.TextField()
    fk_not_in_admin = models.ForeignKey(InAdmin, null=True, on_delete=models.CASCADE)
    model_not_in_admin = models.ForeignKey(
        NotInAdmin, null=True, on_delete=models.CASCADE
    )

    def is_onsale(self):
        return False


class SKU(models.Model):
    name = models.TextField()
    product = models.ForeignKey(Product, on_delete=models.CASCADE)
